#include <conio.h>
#include <stdio.h>
#include <stdlib.h>

#include "registry.hpp"

#include <windows.h>
#include <process.h>
#include "TaskBarApplet.hpp"
#include "TaskBarAppletMain.hpp"
#include "resource.h"

using namespace JetByteTools;

class Applet : public CJBTaskBarApplet
{
public: 
  Applet();
  static void close();

protected:
  void savePrefs();

  virtual LPCTSTR GetTrayTip();

  HICON myIcon;
  virtual HICON GetIcon();

  HMENU hMenu;
  virtual HMENU GetMenu();
  virtual void HandleMenuCommand(int command);
  void updateMenu();
  enum Menu_t {
    Menu3bc,
    Menu378,
    Menu278,
    MenuSep1,
    MenuLED8,
    MenuLED12,
    MenuLED32,
    MenuSep2,
    MenuClose,
  };

  Registry prefs;
};

Applet myApplet;

#define STROBE0 1
#define STROBE1 0
#define AUTOFEED0 2
#define AUTOFEED1 0
#define INIT0 0
#define INIT1 4
#define SELECT0 8
#define SELECT1 0

#define LEDE0 SELECT0
#define LEDE1 SELECT1
#define LEDA0 AUTOFEED0
#define LEDA1 AUTOFEED1
#define LEDB0 INIT0
#define LEDB1 INIT1

enum ledtype_t { LED8, LED12, LED32 };
volatile ledtype_t ledtype=LED8;
volatile int port=0x378;
volatile bool run=true;

void LEDout8(unsigned char data)
{
  _outp(port, data);
}

void LEDout12(unsigned int data)
{
  unsigned char command=0;
  if(data&0x001)
    command|=STROBE1;
  else
    command|=STROBE0;
  if(data&0x200)
    command|=AUTOFEED1;
  else
    command|=AUTOFEED0;
  if(data&0x400)
    command|=INIT1;
  else
    command|=INIT0;
  if(data&0x800)
    command|=SELECT1;
  else
    command|=SELECT0;
  _outp(port, (data>>1)&0xff);
  _outp(port+2, command);
}

void LEDout32(int data, int num)
{
  const unsigned char LEDcommand[]={LEDB0|LEDA0, LEDB0|LEDA1, LEDB1|LEDA0, LEDB1|LEDA1};
  const unsigned char command=LEDcommand[num];
  _outp(port+2, command|LEDE1);
  _outp(port, data);
  _outp(port+2, command|LEDE0); // windows is too fast for our hardware
  _outp(port+2, command|LEDE0);
  _outp(port+2, command|LEDE1);
  _outp(port+2, command|LEDE1);
}

void LEDclear()
{
  int i;
  switch(ledtype)
    {
    default:
    case LED8:
      LEDout8(0);
      _outp(port+2, STROBE0|AUTOFEED0|INIT0|SELECT0);
      break;
    case LED12:
      LEDout12(0);
      break;
    case LED32:
      for(i=0; i<4; i++)
	LEDout32(0, i);
      break;
    }
}

void LED(unsigned int nn)
{
  int i;
  if(nn>100)
    nn=100;
  switch(ledtype)
    {
    default:
    case LED8:
      nn/=12;
      if(nn>8)
       nn=8;
      LEDout8((1<<nn)-1);
      break;
    case LED12:
      nn/=8;
      if(nn>12)
        nn=12;
      LEDout12((1<<nn)-1);
      break;
    case LED32:
      {
        nn/=3;
        if(nn>=32)
          nn=~0;
        else
          nn=(1<<nn)-1;
        static unsigned int old=0;
        const unsigned int nn2=nn;
        for(i=0; i<4; i++)
          {
            unsigned char n=nn&0xff;
            unsigned char o=old&0xff;

            if(n!=o)
              LEDout32(n, i);

            nn>>=8;
            old>>=8;
          }
        old=nn2;
      }
      break;
    }

}

void LEDrun(void *d)
{
  BYTE buf[128];
  // init cpu countmeter service
  try
  {
    Registry r("PerfStats\\StartSrv", HKEY_DYN_DATA);
    DWORD l=sizeof(buf);
    r.get("KERNEL", buf, l);
  }
  catch(...)
  {
    Applet::close();
    return;
  }
  try
  {
    Registry r("PerfStats\\StartStat", HKEY_DYN_DATA);
    DWORD l=sizeof(buf);
    r.get("KERNEL\\CPUUsage", buf, l);
  }
  catch(...)
  {
    Applet::close();
    return;
  }

  // now run it
  try
  {
    LEDclear();
    const double falloff=.9;
    const double falloff1=1.0-falloff;
    double oldcpu=0;
    Registry Stats("PerfStats\\StatData", HKEY_DYN_DATA);
    while(run)
    {
      DWORD l=sizeof(buf);
      Stats.get("KERNEL\\CPUUsage", buf, l);
      oldcpu*=falloff;
      oldcpu+=falloff1*buf[0];
      LED(oldcpu);
      Sleep(50);
    }

    LEDclear();
  }
  catch(...)
  {
    Applet::close();
    return;
  }

  // and stop service again
  try
  {
    Registry r("PerfStats\\StopStat", HKEY_DYN_DATA);
    DWORD l=sizeof(buf);
    r.get("KERNEL\\CPUUsage", buf, l);
  }
  catch(...)
  { }
  try
  {
    Registry r("PerfStats\\StopSrv", HKEY_DYN_DATA);
    DWORD l=sizeof(buf);
    r.get("KERNEL", buf, l);
  }
  catch(...)
  { }
}



Applet::Applet() :
  CJBTaskBarApplet(),
  myIcon(0),
  prefs("SOFTWARE\\cubic\\led")
{
  try
  {
    port=prefs.getInt("port");
    ledtype=static_cast<ledtype_t>(prefs.getInt("ledtype"));
  }
  catch(...)
  { }

  _beginthread(LEDrun, 0, NULL);
}

void Applet::savePrefs()
{
  try
  {
    prefs.set("port", port);
    prefs.set("ledtype", ledtype);
  }
  catch(...)
  { }
}

LPCTSTR Applet::GetTrayTip()
{
  return "LED http://llg.cubic.org/tools";
}

HICON Applet::GetIcon()
{
  if(!myIcon)
  {
    myIcon=LoadIcon(s_hInstance, MAKEINTRESOURCE(IDI_ICON4));
  }
  return myIcon;
}

HMENU Applet::GetMenu()
{
  hMenu = ::CreatePopupMenu();

  MENUITEMINFO item;
  memset(&item, 0, sizeof(MENUITEMINFO));
  item.cbSize = sizeof(MENUITEMINFO);
  item.fMask = MIIM_TYPE | MIIM_ID;
  item.fType = MFT_STRING;

  item.dwTypeData = "LPT0 0x3BC"; item.wID = Menu3bc; ::InsertMenuItem(hMenu, item.wID, TRUE, &item);
  item.dwTypeData = "LPT1 0x378"; item.wID = Menu378; ::InsertMenuItem(hMenu, item.wID, TRUE, &item);
  item.dwTypeData = "LPT2 0x278"; item.wID = Menu278; ::InsertMenuItem(hMenu, item.wID, TRUE, &item);
  item.dwTypeData = "8 Leds"; item.wID = MenuLED8; ::InsertMenuItem(hMenu, item.wID, TRUE, &item);
  item.dwTypeData = "12 Leds"; item.wID = MenuLED12; ::InsertMenuItem(hMenu, item.wID, TRUE, &item);
  item.dwTypeData = "32 Leds"; item.wID = MenuLED32; ::InsertMenuItem(hMenu, item.wID, TRUE, &item);
  item.dwTypeData = "Close"; item.wID = MenuClose; ::InsertMenuItem(hMenu, item.wID, TRUE, &item);

  item.fType = MFT_SEPARATOR;
  item.wID=MenuSep1; ::InsertMenuItem(hMenu, item.wID, true, &item);
  item.wID=MenuSep2; ::InsertMenuItem(hMenu, item.wID, true, &item);

  updateMenu();
  return hMenu;
}

void Applet::HandleMenuCommand(int command)
{
  LEDclear();
  switch (command)
  {
  case MenuClose: run=false; CloseApplet(); return;
  case Menu278: port=0x278; break;
  case Menu378: port=0x378; break;
  case Menu3bc: port=0x3bc; break;
  case MenuLED8:  ledtype=LED8; break;
  case MenuLED12: ledtype=LED12; break;
  case MenuLED32: ledtype=LED32; break;
  }
  updateMenu();
  savePrefs();
}

void Applet::updateMenu()
{
  MENUITEMINFO item;
  memset(&item, 0, sizeof(MENUITEMINFO));
  item.cbSize = sizeof(MENUITEMINFO);
  item.fMask = MIIM_STATE | MIIM_ID;

  item.wID=Menu278; item.fState=(port==0x278)?MFS_CHECKED:MFS_UNCHECKED; SetMenuItemInfo(hMenu, item.wID, true, &item);
  item.wID=Menu378; item.fState=(port==0x378)?MFS_CHECKED:MFS_UNCHECKED; SetMenuItemInfo(hMenu, item.wID, true, &item);
  item.wID=Menu3bc; item.fState=(port==0x3bc)?MFS_CHECKED:MFS_UNCHECKED; SetMenuItemInfo(hMenu, item.wID, true, &item);
  item.wID=MenuLED8;  item.fState=(ledtype==LED8)?MFS_CHECKED:MFS_UNCHECKED;  SetMenuItemInfo(hMenu, item.wID, true, &item);
  item.wID=MenuLED12; item.fState=(ledtype==LED12)?MFS_CHECKED:MFS_UNCHECKED; SetMenuItemInfo(hMenu, item.wID, true, &item);
  item.wID=MenuLED32; item.fState=(ledtype==LED32)?MFS_CHECKED:MFS_UNCHECKED; SetMenuItemInfo(hMenu, item.wID, true, &item);
}

void Applet::close()
{
  SendMessage(Applet::s_hWnd, WM_DESTROY, 0, 0);
}
